/***************************************************************************
 *   Copyright (C) 2011-2013 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

 
#include "sprite.h"
#include "level.h"
#include "../networkengine/filerepository.h"

#include <QGraphicsScene>
#include <QGraphicsSceneMouseEvent>

bool Sprite::leftKeyPressed;
bool Sprite::rightKeyPressed;
bool Sprite::upKeyPressed;
bool Sprite::downKeyPressed;

int Sprite::playerFireSpeed;	//speed of player fire
int Sprite::playerXSpeed;		//speed of player along x (when a key is pressed)
int Sprite::playerYSpeed;		//speed of player along y (when a key is pressed)

QPointF Sprite::mouseInitialCoordinates;	//initial mouse coordinates during a drag-drop
qreal Sprite::spriteInitialX;				//initial sprite X coordinate during a drag-drop
qreal Sprite::spriteInitialY;				//initial sprite Y coordinate during a drag-drop

Level* Sprite::currentLevel;

Sprite::Sprite() {
	movableOutScene = true;
}

//QGraphicsItem flag
void Sprite::setSelectable(bool isSelectable) {
	setFlag(ItemIsSelectable, isSelectable);
}

//QGraphicsItem flag
void Sprite::setMovable(bool isMovable) {
	setFlag(ItemIsMovable, isMovable);
}

//specific to Sprite
void Sprite::setMovableOutScene(bool isMovable) {
	movableOutScene = isMovable;
}

void Sprite::setSpritePixmap(const QString& shortFileName) {
	QPixmap pixmap;		//create a null pixmap
	
	if(!shortFileName.isEmpty()) {
		//result is a null pixmap if the file could not be loaded correctly
		pixmap.load(FileRepository::getAbsoluteFileName(shortFileName));
	}
	
	if(pixmap.isNull()) {
		//file does not exist or invalid format
		//we use the default pixmap embedded in the binary
		pixmap.load(":/images/file_broken.png");
	}
	
	setPixmap(pixmap);	//function of QGraphicsPixmapItem
}

void Sprite::setParameters(SpriteParameter p) {
	//init the pixmap
	QString shortFileName = currentLevel->getPixmapName(p.spriteShape);
	setSpritePixmap(shortFileName);
	
	//initialize the width and height of the sprite
	QRectF rect = boundingRect();
	spriteWidth = rect.width();
	spriteHeight = rect.height();
	
	//set the parameters
	param = p;
	setPos(param.x_pos, param.y_pos);
	setZValue(param.z_pos);
	
	//special case of the player sprite
	if(param.spriteShape==0) {
		//save the values of x, y and fire speed
		playerFireSpeed = param.fireSpeed;
		playerXSpeed = param.x_speed;
		playerYSpeed = param.y_speed;
		
		//initial speeds are 0 (move is triggered by keyboard)
		param.fireSpeed = 0;
		param.x_speed = 0;
		param.y_speed = 0;
	}
	
	//initialize also the other variables
	if(param.fireSpeed!=0) {
		//case of automated fire : set fire_cycle so that we are near the end of the cycle
		fire_cycle = 3*param.fireSpeed/4;
	} else {
		fire_cycle = 0;
	}
	isDestroyed = false;
	explodingCycle = 0;
}

SpriteParameter Sprite::getParameters() {
	if(param.spriteShape==0) {
		//special case of the player sprite
		//the parameters that can be controlled by the players are stored in different variables
		SpriteParameter playerParameters = param;
		playerParameters.fireSpeed = playerFireSpeed;
		playerParameters.x_speed = playerXSpeed;
		playerParameters.y_speed = playerYSpeed;
		return playerParameters;
	} else {
		//normal case, return the current parameters
		return param;
	}
}

//apply a correction on the y_position of the height of the sprite
//it is used to initialize correctly the y starting position of fire that goes in the up direction
//without this correction, we don't know the correct y value to avoid collision with the firing sprite
void Sprite::heightCorrection() {
	param.y_pos -= spriteHeight;
	setPos(param.x_pos, param.y_pos);
}

int Sprite::getSpriteShape() {
	return param.spriteShape;
}

//called when scene.advance() is called (at regular intervals)
//called one time with step=0 and one time with step!=0
//all the necessary computations for collisions are done when step=0
//in step=1, we emit the destroy signals for the sprites that were destroyed
//TODO : this function should be split in more elementary functions
void Sprite::advance(int step) {
	if(step==0) {
		//variation of the position depending on the speed
		qreal x_new = param.x_pos + param.x_speed;
		qreal y_new = param.y_pos + param.y_speed;
		if(x_new < Global::gameViewX) {
			x_new = Global::gameViewX;
			if(param.limitPolicy==SpriteParameter::DestroyWhenOut && !isDestroyed) {
				isDestroyed = true;
			}
		}
		if(x_new > Global::gameViewX + Global::gameViewW - spriteWidth) {
			x_new = Global::gameViewX + Global::gameViewW - spriteWidth;
			if(param.limitPolicy==SpriteParameter::DestroyWhenOut && !isDestroyed) {
				isDestroyed = true;
			}
		}
		if(y_new < Global::gameViewY) {
			if(param.limitPolicy==SpriteParameter::BlockInScreen) {
				y_new = Global::gameViewY;
			} else
			if(param.limitPolicy==SpriteParameter::DestroyWhenOut) {
				//destroy the widget only when it becomes invisible
				//(with a margin of 5 to avoid problems when creating items)
				if(!isDestroyed && y_new < Global::gameViewY - spriteHeight - 5) {
					isDestroyed = true;
				}
			}
		}
		if(y_new > Global::gameViewY + Global::gameViewH - spriteHeight) {
			if(param.limitPolicy==SpriteParameter::BlockInScreen) {
				y_new = Global::gameViewY + Global::gameViewH - spriteHeight;
			} else
			if(param.limitPolicy==SpriteParameter::DestroyWhenOut) {
				//destroy the widget only when it becomes invisible
				//(with a margin of 5 to avoid problems when creating items)
				if(!isDestroyed && y_new > Global::gameViewY + Global::gameViewH + 5) {
					isDestroyed = true;
				}
			}
		}
		
		param.x_pos = x_new;
		param.y_pos = y_new;
		setPos(param.x_pos, param.y_pos);
		
		if(param.fireSpeed !=0) {
			if(fire_cycle == param.fireSpeed) {
				SpriteParameter paramFire = currentLevel->getSpriteDefaultParameters(param.fireIndex);
				qreal margin = 3;	//small margin to avoid collision of the firing sprite with its own fire
				paramFire.x_pos = param.x_pos + 0.5*spriteWidth; //fire at the center of the firing sprite
				if(paramFire.y_speed <= 0) {
					//fire is going up : start from the up of firing sprite
					//fire initial position needs to be moved up from the height of fire
					//it is done later when creating the sprite
					paramFire.y_pos = param.y_pos - margin;
				} else {
					//fire is going down : start from the down of firing sprite
					paramFire.y_pos = param.y_pos + spriteHeight + margin;
				}
				emit fire(paramFire);
				fire_cycle = 0;
			} else {
				fire_cycle++;
			}
		}
		
		//we check only the collisions of items that have a collision policy set to Destroyed
		//These items will set directly items with DestroyedIndirectly policy to destroyed
		//collisions are detected even if exploding cycle is already launched in order to destroy
		//  DestroyedIndirectly items (it means that fire disappears when it collides with an exploding ship)
		if(param.collisionPolicy==SpriteParameter::Destroyed) {
			QList<QGraphicsItem *> collidingItemsList = scene()->collidingItems(this);
			
			//look for something different than background items
			int i;
			int n = collidingItemsList.size();
			for(i=0; i<n; i++) {
				//sprite is killed if it hits anything at the same "depth" (z_pos) as itself
				Sprite * collidingSprite = (Sprite *) (collidingItemsList.at(i));
				if(collidingSprite->param.z_pos == param.z_pos) {
					//start the exploding cycle only if it was not started already
					if(explodingCycle==0) {
						//decrease the life points from the number of life points of the colliding sprite
						//in case of fire, life points describe the destruction power
						param.lifePoints -= collidingSprite->param.lifePoints;
						
						//sprite starts to explode when it has no more life points
						if(param.lifePoints<=0) {
							emit askSound(1);
							explodingCycle = 1;
							param.fireSpeed = 0;	//stop any fire from the exploding ship
							fire_cycle = 0;
						} else {
							emit askSound(2);
						}
					}
					//also destroy immediately the sprite that collided (if it is fire)
					if(collidingSprite->param.collisionPolicy==SpriteParameter::DestroyedIndirectly) {
						collidingSprite->isDestroyed=true;
					}
				}
			}
		}
		
		if(explodingCycle!=0) {
			int animDelay = currentLevel->getSpriteKindParameter(param.spriteShape).animationSpeed;
			if(explodingCycle == animDelay * currentLevel->getSpriteKindParameter(param.spriteShape).animationLength ) {
				//end of the explosing cycle, delete the ship from the scene
				isDestroyed=true;
			} else
			if(explodingCycle%animDelay == 0) {
				QString animFileName = currentLevel->levelParam.animationFileName(param.spriteShape, (explodingCycle/animDelay)+1);
				setSpritePixmap(animFileName);
			}
			explodingCycle++;
		}
	} else {
		//called after the first step
		if(isDestroyed) {
			emit destroy();
		}
	}
}

//return true if the key is accepted by the sprite
//we remember the last pressed key in x and y direction, so that we don't react to the release of the other key
//it gives better control because most players begin to press a key before releasing the previous one
bool Sprite::treatKey(int pressedKey, bool isPressed) {
	switch (pressedKey) {
		case Qt::Key_Up:
			upKeyPressed = isPressed;
			if(isPressed) {
				param.y_speed = -playerYSpeed;
			} else {
				//up key was just released, go down if down key is pressed
				if(downKeyPressed) {
					param.y_speed = playerYSpeed;
				} else{
					param.y_speed = 0;
				}
			}
			break;
		case Qt::Key_Down:
			downKeyPressed = isPressed;
			if(isPressed) {
				param.y_speed = playerYSpeed;
			} else {
				//down key was just released, go up if up key is pressed
				if(upKeyPressed) {
					param.y_speed = -playerYSpeed;
				} else {
					param.y_speed = 0;
				}
			}
			break;
		case Qt::Key_Left:
			leftKeyPressed = isPressed;
			if(isPressed) {
				param.x_speed = -playerXSpeed;
			} else {
				//left key was just released, go right if right key is pressed
				if(rightKeyPressed) {
					param.x_speed = playerXSpeed;
				} else {
					param.x_speed = 0;
				}
			}
			break;
		case Qt::Key_Right:
			rightKeyPressed = isPressed;
			if(isPressed) {
				param.x_speed = playerXSpeed;
			} else {
				//right key was just released, go left if left key is pressed
				if(leftKeyPressed) {
					param.x_speed = -playerXSpeed;
				} else {
					param.x_speed = 0;
				}
			}
			break;
		case Qt::Key_Space:
			if(isPressed) {
				param.fireSpeed = playerFireSpeed;
				fire_cycle = playerFireSpeed;			//same value to shoot immediatly
			} else {
				param.fireSpeed = 0;
			}
			break;
		default:
			return false;
	}
	return true;
}


//We reimplement QGraphicsItem::mousePressEvent to catch the mouse initial position
//This will be used to compute the move coordinates
void Sprite::mousePressEvent(QGraphicsSceneMouseEvent *mouseEvent) {
	mouseInitialCoordinates = mouseEvent->scenePos();
	spriteInitialX = param.x_pos;
	spriteInitialY = param.y_pos;
	emit spriteSelected(getParameters());			//signal the parameters of the sprite (used in edit mode)
	QGraphicsItem::mousePressEvent(mouseEvent);		//call the default implementation
}

//The default function QGraphicsItem::mouseMoveEvent doesn't allow us to know simply the coordinates change
// -> we reimplement the mouseMoveEvent to catch the change
// it also allows us to perform some operations (for example : forbid the movement out of the scene)
//note : the current version only takes into account the last selected item
//the code is directly inspired from QGraphicsItem::mouseMoveEvent
void Sprite::mouseMoveEvent(QGraphicsSceneMouseEvent *mouseEvent) {
	if ((mouseEvent->buttons() & Qt::LeftButton) && (flags() & ItemIsMovable)) {
		//compute the new position coordinates from the current mouse position
		QPointF translation = mouseEvent->scenePos() - mouseInitialCoordinates;
		param.x_pos = spriteInitialX + translation.x();
		param.y_pos = spriteInitialY + translation.y();
		
		//take into account the scene limit
		param.enforceEditSceneLimits(spriteWidth, spriteHeight);
		
		//set the new position
		setPos(param.x_pos, param.y_pos);
		emit spriteMoved(param);
	} else {
		mouseEvent->ignore();
	}
}

/*
void Sprite::mouseReleaseEvent(QGraphicsSceneMouseEvent *mouseEvent) {
	QGraphicsItem::mouseReleaseEvent(mouseEvent);		//call the default implementation
}
*/
